"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const source_map_support_1 = require("source-map-support");
const chalk = require("chalk");
const cli_highlight_1 = require("cli-highlight");
const Helper_1 = require("../Helper");
const path_1 = require("path");
const util_1 = require("util");
class Logger {
    constructor(nodeId, name, minLevel = 0) {
        this.nodeId = nodeId;
        this.name = name;
        this.logLevels = {
            0: "silly",
            1: "trace",
            2: "debug",
            3: "info",
            4: "warn",
            5: "error",
        };
        this.logLevelsColors = {
            0: "#B0B0B0",
            1: "#FFFFFF",
            2: "#63C462",
            3: "#2020C0",
            4: "#CE8743",
            5: "#CD444C",
        };
        this.ignoreStackLevels = 3;
        this.doOverwriteConsole = false;
        this.logAsJson = false;
        this.errorToJsonHelper();
        // TODO: catch all errors & exceptions
        // x Log as json
        // x remove ms
        // transport (inkl. min level)
        // config
        // override console.log
        // x print out logger name, if set
        if (this.doOverwriteConsole) {
            this.overwriteConsole();
        }
    }
    silly(...args) {
        return this.handleLog.apply(this, [0, args]);
    }
    trace(...args) {
        return this.handleLog.apply(this, [1, args, true]);
    }
    debug(...args) {
        return this.handleLog.apply(this, [2, args]);
    }
    info(...args) {
        return this.handleLog.apply(this, [3, args]);
    }
    warn(...args) {
        return this.handleLog.apply(this, [4, args]);
    }
    error(...args) {
        return this.handleLog.apply(this, [5, args]);
    }
    handleLog(logLevel, logArguments, doExposeStack = true) {
        const logObject = this.buildLogObject(logLevel, logArguments, doExposeStack);
        if (!this.logAsJson) {
            this.printPrettyLog(logObject);
        }
        else {
            this.printJsonLog(logObject);
        }
        return logObject;
    }
    buildLogObject(logLevel, logArguments, doExposeStack = true) {
        var _a;
        const callSites = this.getCallSites();
        const relevantCallSites = callSites.splice(this.ignoreStackLevels);
        const stackFrame = this.wrapCallSiteOrIgnore(relevantCallSites[0]);
        const stackFrameObject = this.toStackFrameObject(stackFrame);
        const logObject = {
            loggerName: (_a = this.name) !== null && _a !== void 0 ? _a : "",
            date: new Date(),
            logLevel: logLevel,
            logLevelName: this.logLevels[logLevel],
            filePath: stackFrameObject.filePath,
            fullFilePath: stackFrameObject.fullFilePath,
            fileName: stackFrameObject.fileName,
            lineNumber: stackFrameObject.lineNumber,
            columnNumber: stackFrameObject.columnNumber,
            isConstructor: stackFrameObject.isConstructor,
            functionName: stackFrameObject.functionName,
            typeName: stackFrameObject.typeName,
            methodName: stackFrameObject.methodName,
            argumentsArray: [],
        };
        logArguments.forEach((arg) => {
            if (typeof arg === "object" && !util_1.types.isNativeError(arg)) {
                logObject.argumentsArray.push(JSON.parse(JSON.stringify(arg)));
            }
            else if (typeof arg === "object" && util_1.types.isNativeError(arg)) {
                const errorStack = this.getCallSites(arg);
                const errorObject = JSON.parse(JSON.stringify(arg));
                errorObject.isError = true;
                errorObject.stack = this.toStackObjectArray(errorStack);
                logObject.argumentsArray.push(errorObject);
            }
            else {
                logObject.argumentsArray.push(util_1.format(arg));
            }
        });
        if (doExposeStack) {
            logObject.stack = this.toStackObjectArray(relevantCallSites);
        }
        return logObject;
    }
    toStackObjectArray(jsStack) {
        let prettyStack = Object.values(jsStack).reduce((iPrettyStack, stackFrame) => {
            iPrettyStack.push(this.toStackFrameObject(this.wrapCallSiteOrIgnore(stackFrame)));
            return iPrettyStack;
        }, []);
        return prettyStack;
    }
    toStackFrameObject(stackFrame) {
        var _a, _b;
        const filePath = stackFrame.getFileName();
        return {
            filePath: (_a = Helper_1.Helper.cleanUpFilePath(filePath)) !== null && _a !== void 0 ? _a : "",
            fullFilePath: filePath !== null && filePath !== void 0 ? filePath : "",
            fileName: path_1.basename((_b = stackFrame.getFileName()) !== null && _b !== void 0 ? _b : ""),
            lineNumber: stackFrame.getLineNumber(),
            columnNumber: stackFrame.getColumnNumber(),
            isConstructor: stackFrame.isConstructor(),
            functionName: stackFrame.getFunctionName(),
            typeName: stackFrame.getTypeName(),
            methodName: stackFrame.getMethodName(),
        };
    }
    printPrettyLog(logObject) {
        // only errors should go to stdErr
        const std = logObject.logLevel < 5 ? process.stdout : process.stderr;
        const nowStr = logObject.date
            .toISOString()
            .replace("T", " ")
            .replace("Z", "");
        const hexColor = this.logLevelsColors[logObject.logLevel];
        std.write(chalk `{grey ${nowStr}}\t`);
        std.write(chalk.hex(hexColor).bold(` ${logObject.logLevelName.toUpperCase()}\t`));
        const functionName = logObject.isConstructor
            ? `${logObject.typeName}.constructor`
            : logObject.methodName != null
                ? `${logObject.typeName}.${logObject.methodName}`
                : `${logObject.functionName}`;
        std.write(chalk.gray(`[${logObject.loggerName}@${this.nodeId} ${logObject.filePath}:${logObject.lineNumber} ${functionName}]\t`));
        logObject.argumentsArray.forEach((arg) => {
            if (typeof arg === "object" && !arg.isError) {
                std.write("\n" +
                    cli_highlight_1.highlight(JSON.stringify(arg, null, 2), { language: "JSON" }) +
                    " ");
            }
            else if (typeof arg === "object" && arg.isError) {
                std.write(util_1.format(chalk `\n{whiteBright.bgRed.bold ${arg.name}}{grey :} ${util_1.format(arg.message)}\n`));
                this.printPrettyStack(std, arg.stack);
            }
            else {
                std.write(util_1.format(arg + " "));
            }
        });
        std.write("\n");
        if (logObject.stack != null) {
            std.write("log stack:\n");
            this.printPrettyStack(std, logObject.stack);
        }
    }
    printPrettyStack(std, stackObjectArray) {
        std.write("\n");
        Object.values(stackObjectArray).forEach((stackObject) => {
            var _a;
            std.write(chalk `    {grey •} {yellowBright ${stackObject.fileName}}{grey :}{yellow ${stackObject.lineNumber}} {white ${(_a = stackObject.functionName) !== null && _a !== void 0 ? _a : "<anonumous>"}}`);
            std.write("\n    ");
            std.write(path_1.normalize(chalk `{grey ${stackObject.fullFilePath}:${stackObject.lineNumber}:${stackObject.columnNumber}}`));
            std.write("\n\n");
        });
    }
    printJsonLog(logObject) {
        // only errors should go to stdErr
        const std = logObject.logLevel < 5 ? process.stdout : process.stderr;
        std.write(cli_highlight_1.highlight(JSON.stringify(logObject)) + "\n");
    }
    wrapCallSiteOrIgnore(callSiteFrame) {
        try {
            return source_map_support_1.wrapCallSite(callSiteFrame);
        }
        catch {
            return callSiteFrame;
        }
    }
    getCallSites(error) {
        const _prepareStackTrace = Error.prepareStackTrace;
        Error.prepareStackTrace = (_, stack) => stack;
        const stack = error == null ? new Error().stack.slice(1) : error.stack;
        Error.prepareStackTrace = _prepareStackTrace;
        return stack;
    }
    errorToJsonHelper() {
        if (!("toJSON" in Error.prototype))
            Object.defineProperty(Error.prototype, "toJSON", {
                value: function () {
                    return Object.getOwnPropertyNames(this).reduce((alt, key) => {
                        alt[key] = this[key];
                        return alt;
                    }, {});
                },
                configurable: true,
                writable: true,
            });
    }
    overwriteConsole() {
        const $this = this;
        ["log", "debug", "info", "warn", "trace", "error"].forEach(function (name) {
            /*console[name] = (...args: any[]) => {
              // force console.log!!
              return $this.handleLog.apply($this, [5, args]);
            };*/
        });
    }
}
exports.Logger = Logger;
//# sourceMappingURL=index.js.map